/*
Copyright (c) 2006, CARDON DE LICHTBUER Rodolphe
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list 
of conditions and the following disclaimer. 

2. Redistributions in binary form must reproduce the above copyright notice, this 
list of conditions and the following disclaimer in the documentation and/or 
other materials provided with the distribution. 

3. The name of the author or contributors may not be used to endorse or promote 
products derived from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED 
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH 
DAMAGE.
 */

package be.cardon.sign;

/**Sign one file.
 * USAGE : java SignOneFile [file]
 * This tool :
 *  1. Takes a file as data to sign.
 *  2. Let you choice your private key (MS Crypto, file with Java KS, file with PKCS#12, PKCS#11).
 *  3. Get the certificate chain of your private key.
 *  4. Create an detached XML signature. 
 *  5. Store this signature with name = filename to sign + ".signature.xml"
 * @author Rodolphe
 */

import java.security.PrivateKey;
import java.security.KeyStore;
import java.security.cert.Certificate;
import java.security.cert.X509Certificate;
import java.util.ArrayList;
import java.util.List;
import java.util.Iterator;
import java.util.Enumeration;
import be.cardon.cryptoapi.provider.*;
import java.io.*;
import be.cardon.utils.Debug;
import java.net.URI;

public class SignFiles {
    
    /** Creates a new instance of SignOneFile */
    public SignFiles() {
    }
    /**
     * @param args the command line arguments
     */
    public static void main(String[] args) {
        try{
            
        Options options = readOptions(args);

        //debugging
        be.cardon.utils.Debug.setDebug(options.debug);
        
        //files to sign
        if(options.files2sign.size()<1){
            System.err.println("min one file to sign required.");
            System.exit(1);
        }
        File[] files2sign = new File[options.files2sign.size()];
        for(int i=0;i<options.files2sign.size();i++){
            files2sign[i] = new File(options.files2sign.get(i));
            Debug.println("File to sign = "+files2sign[i].getAbsolutePath());
        }
        
        //signature file
        String signatureFilename;
        if(options.sigFilePath!=null){
            signatureFilename = options.sigFilePath;
        }else{
            //relative to first file
            signatureFilename = options.files2sign.get(0) + ".sig.xml";
        }
        File signatureFile = new File(signatureFilename);
        Debug.println("Signature output file = "+signatureFile.getAbsolutePath());
        
        //key and certificates
        KeyAndChain keyAndChain = selectKey();
        System.out.println("Selected key = "+keyAndChain.getKey());
        System.out.println("chain :");
        be.cardon.utils.Dumper.dumpChain(keyAndChain.getChain(),System.out);
        List<X509Certificate> X509certs  = new java.util.ArrayList<X509Certificate>();
        Certificate[] chain = keyAndChain.getChain();
        for(int i=0; i<chain.length;i++){
            X509certs.add((X509Certificate)chain[i]);
        }
        
        //baseURI
        URI baseURI;
        if(options.baseDir==null){
            baseURI = signatureFile.toURI();
        }else{
            baseURI = signatureFile.toURI().resolve(options.baseDir);
        }
        
        OutputStream os = new BufferedOutputStream(new FileOutputStream(signatureFile));
        Writer writer = be.cardon.sign.xmldsig.XMLSignerEngine.StandardOutputWriter;
        
        
        be.cardon.sign.xmldsig.XMLSignerEngine.Sign2XML(
            files2sign,
            baseURI,
            keyAndChain.getKey(), 
            X509certs,
            java.util.Calendar.getInstance(),
            os,
            writer);
                
        }catch(Exception e){
            e.printStackTrace();
        }
        
    }
    
    public static KeyAndChain selectKey(){
        String menu = "Private key selection menu.\n" +
                "[1] Microsoft KeyStore\n" +
                "[2] JKS file\n" +
                "[3] PKCS#12 file\n";
               // "[4] PKCS#11 DLL\n";
        int menuChoice = KeyboardChoice(menu, 1, 3, true);
        switch(menuChoice){
            case -1 : return null;
            case 1 : return selectMSKey();
            case 2 : return selectFileKey("JKS");
            case 3 : return selectFileKey("PKCS12");
            //case 4 : return selectPKCS11Key();
        }
        System.err.println("BUG : should not come here !");
        return null;
        
    }
    
    public static KeyAndChain selectMSKey(){
        java.security.Security.insertProviderAt(new CryptoAPIProvider(), 1);
        try{
            KeyStore msStore = KeyStore.getInstance("CryptoAPI");
            byte[] namebyte = "my".getBytes("UTF-8");
            InputStream is = new ByteArrayInputStream(namebyte);
            msStore.load(is,null);
            return selectKeyInStore(msStore);
        }catch(Exception e){
            e.printStackTrace(System.out);
            return null;
        }
    }
    
    public static KeyAndChain selectFileKey(String keyStoreType){
        KeyStore ksStore;
        try{
        ksStore = KeyStore.getInstance(keyStoreType);
        }catch(Exception e){
            e.printStackTrace(System.out);
            return null;
        }
        while(true){
            try{
                String storeFileName = be.cardon.utils.Keyboard.input("KeyStore filename (or enter to abort):");
                if(storeFileName.length()==0){return null;}
                new FileInputStream(storeFileName); //test
                while(true){
                    try{
                       String kspass = be.cardon.utils.Keyboard.input("KeyStore Password:");
                       ksStore.load(new FileInputStream(storeFileName),kspass.toCharArray());
                       break;
                    }catch(Exception e){
                        e.printStackTrace(System.out);
                    }
                }
                break;
            }catch (Exception e){
                e.printStackTrace(System.out);
            }
        }
        return selectKeyInStore(ksStore);
    }
    
    public static KeyAndChain selectKeyInStore(KeyStore store){
        // make menu and key list
        try{
        String menu = "Private key selection :\n";
        List<String> keyList = new ArrayList<String>();
        int index = 0;
        for(Enumeration<String> it = store.aliases(); it.hasMoreElements();){
            String alias = it.nextElement();
            if(store.isKeyEntry(alias)){
                menu = menu + "[" + index + "] " + alias + "\n";
                keyList.add(alias);
                index++;
            }
            else{Debug.println("not key entry : "+alias);}
        }
        int max = index;
        if(max == 0){
            System.out.println("no key.");
            return null;
        }
        int choice = KeyboardChoice(menu, 0,max, true);
        if(choice == -1){
            return null;
        }
        PrivateKey mykey;
        String selectAlias = keyList.get(choice);
        while(true){
            try{
                String password = be.cardon.utils.Keyboard.input("Password:");
                mykey = (PrivateKey)store.getKey(selectAlias, password.toCharArray());
                break;
            }catch(Exception e){
                e.printStackTrace(System.out);
            }     
        }
        Certificate[] chain = store.getCertificateChain(selectAlias);
        return new KeyAndChain(mykey, chain);
        }catch(Exception e){
            e.printStackTrace(System.out);
            return null;
        }
    }
    
    
    /**return the choice or -1 if cancelled (if mayCance is true).*/
    public static int KeyboardChoice(String menu, int min, int max, boolean mayCancel){
        boolean CorrectChoice = false;
        Integer inputInt=0;
        while(!CorrectChoice){
            System.out.println(menu);
            String input;
            try{
            input = be.cardon.utils.Keyboard.input("Your choice (or enter to cancel)");
            }catch(Exception e){
                e.printStackTrace(System.out); 
                return -1;
            }
            try{
                if(input.length() == 0 & mayCancel == true){
                    return -1;
                }                
                inputInt = Integer.decode(input);
                if(inputInt < min | inputInt > max){
                    throw new Exception("not in valid range.");
                }
                break;
            }catch(Exception e){
                System.out.println("Incorrect choice. Try again.");
            }
        }
        return inputInt;
    }
    
    public static void printUsage(){
        
        System.out.println("SignOnFile USAGE:");
        System.out.println(" SignOnFile [options] file1 [file2 ...]  ");
        System.out.println("");
        System.out.println("Options:");
        System.out.println("  -v : verbose (debug)");
        System.out.println("  -s file : signature file");
        System.out.println("  -b : base directory (signed files)");
    }

    private static class Options{
 
        public Options(){
            this.files2sign = new ArrayList<String>();
        }
        public boolean debug = false;
        public String file2sign;
        public String sigFilePath=null; // null if undefined
        public List<String> files2sign; 
        public String baseDir=null;//null if undefined
    }
    
    private static Options readOptions(String[] args){
        Options options = new Options();

        if(args.length<1){
            printUsage();
            System.exit(0);
        }
        int c=0;
        while(c < args.length){
            String opt = args[c];
            if(opt.equalsIgnoreCase("-v")){
                options.debug = true;
                c++;
                continue;
            }
            if(opt.equalsIgnoreCase("-s")){
                options.sigFilePath= args[c+1];
                c+=2;
                continue;
            }
            if(opt.equalsIgnoreCase("-b")){
                options.baseDir = args[c+1];
                c+=2;
                continue;
            }
            else{
              if(opt.substring(0,1).equals("-")){
                  System.err.println("Unknown option: "+opt);
                  System.exit(1);
              }
              options.files2sign.add(opt);
              c+=1;
            }
        }
        return options;
    }
        
    public static class KeyAndChain {

        PrivateKey key;
        Certificate[] chain;


        /** Creates a new instance of KeyAndChain */
        public KeyAndChain(PrivateKey key, Certificate[] chain) {
            this.key = key;
            this.chain = chain;
        }

        public PrivateKey getKey(){
            return this.key;
        }

        public Certificate[] getChain(){
            return this.chain;
        }
     }
    
}

