/*
Copyright (c) 2006, CARDON DE LICHTBUER Rodolphe
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list 
of conditions and the following disclaimer. 

2. Redistributions in binary form must reproduce the above copyright notice, this 
list of conditions and the following disclaimer in the documentation and/or 
other materials provided with the distribution. 

3. The name of the author or contributors may not be used to endorse or promote 
products derived from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED 
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH 
DAMAGE.
 */

package be.cardon.sign;

import java.io.*;
import be.cardon.sign.xmldsig.*;
import java.util.*;
import javax.xml.crypto.dsig.*;
import java.security.*;
import java.security.cert.*;
import be.cardon.utils.*;
import java.net.URI;
/**
 *
 * @author Rodolphe
 */
public class VerifySignature {
    
    /** Creates a new instance of VerifOneFile */
    public VerifySignature() {
    }
    
    private static Options readOptions(String[] args){
        Options options = new Options();
        if(args.length<1){
            printUsage();
            System.exit(0);
        }
         options.sigFilePath = args[0];
        int c=1;
        while(c < args.length){
            String opt = args[c];
            if(opt.equalsIgnoreCase("-v")){
                options.debug = true;
                c++;
                continue;
            }
            if(opt.equalsIgnoreCase("-b")){
                options.baseDir = args[c+1];
                c+=2;
                continue;
            }
            if(opt.equalsIgnoreCase("-r")){
                options.revocationCheck = true;
                c+=1;
                continue;
            }
            System.err.println("Unknown param : "+opt);
            System.exit(1);
            
        }
        return options;
    }
    
    public static void printUsage(){
        
        System.out.println("VerifySignature USAGE:");
        System.out.println(" VerifySignature signatureFile [options]");
        System.out.println("");
        System.out.println("Options:");
        System.out.println("  -v : verbose (debug)");
        System.out.println("  -b : base directory (signed files), relative to sig file");
        System.out.println("  -r : revocation check (disabled by default)");
        
        
    }
    /**
     * @param args the command line arguments
     */
    public static void main(String[] args) {
        
        boolean ChainValid = false;
      
        try{
       
        //options setting
        Options options = readOptions(args);
        String SignatureFilename = options.sigFilePath;
        be.cardon.utils.Debug.setDebug(options.debug);
        File signFile = new File(SignatureFilename);
        if(!signFile.canRead()){
            throw new IOException("cannot read file !");
        }

        File baseDir=null;
        if(options.baseDir!=null){
            URI baseDirURI = signFile.toURI().resolve(options.baseDir);
            baseDir = new File(baseDirURI.getPath());
        }
        
     //// CORE XML VALIDITY
        SignatureStatus ss = XMLSignerEngine.VerifierOneFileXMLSignature(signFile, baseDir);
        
        //DUMP status
        System.out.println("+ Core verifification = "+ss.coreValid);
        System.out.println("+ Signature verification = "+ss.signatureValid);
        System.out.println("+ References :");
        for(Iterator<ReferenceStatus> it = ss.refsStatus.iterator();it.hasNext();){
            ReferenceStatus rs = it.next();
            System.out.println("  - Reference URI : "+rs.RefURI);
            System.out.println("    - type : "+rs.refType);
            System.out.println("    - digest validity : "+rs.RefValidDigest);
            System.out.println("    - transformations");
            for(Iterator<Transform> itTransf=rs.refTranfs.iterator(); itTransf.hasNext();){
                Transform transf = itTransf.next();
                System.out.println("       - Transformation : "+transf.getAlgorithm());
            }
        }
        
     /// CERTIFICATE VALIDITY
        //use MS Keystore to verify...
        System.out.println("+ Using Windows Trusted certificates.");
        Provider msProvider = new be.cardon.cryptoapi.provider.CryptoAPIProvider();
        KeyStore rootKS = KeyStore.getInstance("CryptoAPI", msProvider);
        rootKS.load(new ByteArrayInputStream("root".getBytes("UTF-8")),null);
        
        KeyStore caKS = KeyStore.getInstance("CryptoAPI", msProvider);
        caKS.load(new ByteArrayInputStream("ca".getBytes("UTF-8")),null);
        
       
        
        X509CertSelector CertSelector  = new X509CertSelector();
        
        CertSelector.setCertificate(ss.signingCertificate); 
        Debug.println("End user certificate =" +ss.signingCertificate);
        

        //create set of trusted certificate
        Set tas = new java.util.HashSet<TrustAnchor>();
        
        for(Enumeration<String> it = rootKS.aliases() ; it.hasMoreElements();){
            String alias = it.nextElement();

            if(!rootKS.isCertificateEntry(alias)){
                continue;
            }
            X509Certificate cert = (X509Certificate)rootKS.getCertificate(alias);
            Debug.println("Trusted alias : "+cert.getSubjectDN());
            TrustAnchor ta = new TrustAnchor(cert,null);
            tas.add(ta);
        }
        
      
        
       
     
        PKIXBuilderParameters pkixPara = new PKIXBuilderParameters(tas, CertSelector);
       
        
        // revocation check setting
        pkixPara.setRevocationEnabled(options.revocationCheck);
        System.setProperty("com.sun.security.enableCRLDP", "true");
        
        for(Iterator<X509Certificate> it = ss.certs.iterator();it.hasNext();){
            Debug.println("CERT:"+it.next().toString());
        }
        CollectionCertStoreParameters ccsp = new CollectionCertStoreParameters(ss.certs);
        CertStore cs = CertStore.getInstance("Collection", ccsp);
        
        pkixPara.addCertStore(cs);
        pkixPara.setRevocationEnabled(false); //don't user revocation check fot path builder
        PKIXCertPathBuilderResult pcpbr = null;
        
       try{
           CertPathBuilder cpb = CertPathBuilder.getInstance("PKIX");
           

           pcpbr = (PKIXCertPathBuilderResult)cpb.build(pkixPara);

       }catch(Exception e2){
             if(Debug.getEnabled()){
                e2.printStackTrace();
             }
             System.out.println("Unable to find a chain with trusted root.");
        }
        if(pcpbr!=null){

        CertPath path = pcpbr.getCertPath();
        X509Certificate rootCert = pcpbr.getTrustAnchor().getTrustedCert();
        
        //print path
        System.out.println("+ Chain (end certificate first):");
        Integer count = new Integer(0);
        for(Iterator it = path.getCertificates().iterator(); it.hasNext();){
            X509Certificate cert = (X509Certificate)it.next();
            String subject = cert.getSubjectDN().toString();
            System.out.println("  ["+count+"] : "+subject);
            count++;
        }
        System.out.println("  ["+count+"] : "+rootCert.getSubjectDN().toString());
        
        CertPathValidator cpv = CertPathValidator.getInstance("PKIX");
        PKIXParameters params = new PKIXParameters(tas);
        
         // revocation check setting
        java.security.Security.setProperty("ocsp.enable", "true");
        //java.security.Security.setProperty("ocsp.responderURL", "http://ocsp.eid.belgium.be");
        params.setRevocationEnabled(options.revocationCheck);
        System.setProperty("com.sun.security.enableCRLDP", "true");
       

            try{

            CertPathValidatorResult cpvr = cpv.validate(path, params);
            ChainValid = true;
            System.out.println("+ PKIX Cert path validation revocation enabled: "+options.revocationCheck);
            System.out.println("+ PKIX Cert path validation OK");
            if(cpvr instanceof PKIXCertPathValidatorResult){
                PKIXCertPathValidatorResult pcpvr = (PKIXCertPathValidatorResult)cpvr;
                Debug.println("PKIX Cert path validator result:"+pcpvr.toString());

            }
            }catch(Exception e2){
                e2.printStackTrace();
            }
        
        }
        if(ChainValid & ss.coreValid){
            System.out.println("===> SIGNATURE VALID");
        }else{
            System.out.println("===> SIGNATURE NOT VALID !");            
        }
        
        
        }catch(Exception e){
            e.printStackTrace();
        }
    }
    
    private static class Options{
 
        public Options(){

        }
        public boolean debug = false;
        public String sigFilePath;
        public String baseDir=null;
        public boolean revocationCheck = false;


}

    
}

