/*
Copyright (c) 2006, CARDON DE LICHTBUER Rodolphe
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list 
of conditions and the following disclaimer. 

2. Redistributions in binary form must reproduce the above copyright notice, this 
list of conditions and the following disclaimer in the documentation and/or 
other materials provided with the distribution. 

3. The name of the author or contributors may not be used to endorse or promote 
products derived from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED 
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH 
DAMAGE.
 */

package be.cardon.sign.communicator;

import java.util.List;
import java.util.ArrayList;
import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Set;
import java.io.*;
import java.net.*;
import be.cardon.utils.Debug;
import be.cardon.sign.applet.*;
import be.cardon.sign.*;
import java.security.MessageDigest;
import java.security.Provider;
import java.security.DigestInputStream;



/**Download a file
 *The server MUST returns the size of the file in HTTP header, otherwize
 *this class generates a exception and display.
 *This size is used to detect a download interruption.
 * @author Rodolphe
 */
public class DownloaderFile {
    
    public final static int STATUS_NOT_DOWNLOADED = 0;
    public final static int STATUS_CONNECTING = 1; // 0 octets downloaded
    public final static int STATUS_DOWNLOADING = 2; // > 0 octets downloaded
    public final static int STATUS_DOWNLOADED = 3;
    public final static int STATUS_ERROR = -1;

    private int inputBufferSize= 1000;
    private int outputBufferSize= 100000;
    private int internalBufferSize = 10000;
    
    private URL sourceURL;
    private File relativePath;
    private static File defaultTargetDirectory;
    private File targetDirectory;
    private long size; //may be -1
    private int status;
    private boolean interrupt=false;
    private long counter=0;
    private java.util.Map<String,MessageDigest> messageDigestMap=new HashMap<String, MessageDigest>();
    private java.util.Map<String,byte[]> digestsMap;
    
    /** Creates a new instance of DownloaderFile, without InputStream
     *size may be -1 if undefined.
     */

    //size may be -1;
    public DownloaderFile(URL sourceURL, File relativePath, long size) {
        this.sourceURL = sourceURL;
        this.relativePath = relativePath;
        this.size = size;
    }

    public DownloaderFile(URL sourceURL, File relativePath, long size, File targetDirectory) {
        this.sourceURL = sourceURL;
        this.relativePath = relativePath;
        this.size = size;
        this.targetDirectory = targetDirectory;
    }
    /**add digest algorithm. Provider may be null.*/
    public void addDigest(String alg, Provider prov) 
    throws java.security.NoSuchAlgorithmException{
        if(status!=STATUS_NOT_DOWNLOADED && status!=STATUS_ERROR){
            throw new IllegalArgumentException("Invalid call when downloading.");
        }
        if(!messageDigestMap.containsKey(alg)){
            MessageDigest md;
            if(prov==null){
                md = MessageDigest.getInstance(alg);
            }else{
                md = MessageDigest.getInstance(alg, prov);
            }
            messageDigestMap.put(alg, md);
        }
    }
    public byte[] getDigest(String alg){
        if(status!=STATUS_DOWNLOADED){
             throw new IllegalArgumentException("Invalid call if download not ended."); 
        }
        if(!digestsMap.containsKey(alg)){
             throw new IllegalArgumentException("Algorithm not registered before downloading.");
        }
        return digestsMap.get(alg);
    }
    
    public static void setDefaultTargetDirectory(File dir){
        defaultTargetDirectory = dir;
    }
    public static File getDefaultTargetDirectory(File dir){
       return defaultTargetDirectory;
    }
    
    public void setLocalTargetDirectory(File dir){
        this.targetDirectory = dir;
    }
    public URL getSourceURL(){
       return sourceURL;
    };
    public File getRelativePath(){
        return relativePath;
    }
    public File getLocalAbsoluteFile(){
        if(targetDirectory==null && defaultTargetDirectory == null){
            throw new NullPointerException("Call first set(Default)LocalTargetDirectory !");
        }
        if(targetDirectory==null){
            return new File(this.defaultTargetDirectory, relativePath.toString());
        }else{
            return new File(this.targetDirectory, relativePath.toString());
        } 
    }

    public synchronized int getStatus(){
        return status;
    };
    
    private synchronized void setStatus(int status){
        this.status = status;
    };
    
    
    /**Returns the number of remaining octets already downloaded.*/
    public synchronized long getDownloadedSize(){
        return counter;
    }
    
    public synchronized void setCounter(long counter){
        this.counter = counter;
    }
    
    
    /**Returns the number of remaining octets to download.
     *Returns -1 if undetermined (size unknown)
     */
    public long getRemainSize(){
        if(size==-1){
            return -1;
        }
        return size - getDownloadedSize();
    }
    
    /**Returns the total file size.
     * or -1 if undetermined
     */
    public long getSize(){
        return this.size;
    }
    
    /**Interrupt the download.*/
    public synchronized void interrupt(){
        if(status !=STATUS_CONNECTING && status != STATUS_DOWNLOADING){
            System.err.println("DownloaderFile: Interrupt: not downloading");
            return;
        }
        this.interrupt = true;
    }
    
    /**Downloads the file and returns when the file is downloaded, or on error.*/
    public void download() throws IOException, InterruptedException{
        this.interrupt = false;
        CounterInputStream cis;

        //prepare local directory
        File parentDir = this.getLocalAbsoluteFile().getParentFile();
        if(parentDir!=null && !parentDir.exists()){
            if(!parentDir.mkdirs()){
                throw new IOException("DownloaderFile: Unable to create new directory: "+parentDir);            
            }
        }   
        
        //open connection
        setStatus(STATUS_CONNECTING);
                URLConnection con = sourceURL.openConnection();
        int sizeFromServer = con.getContentLength();
        Debug.println("DownloaderFile: sizeFromServer=" +sizeFromServer);
        if(sizeFromServer==-1){
            BufferedReader reader = new BufferedReader(new InputStreamReader(con.getInputStream()));
            String message = reader.readLine();
            reader.close();
            System.err.println("Connection error or server didn't return " +
                    "content length: \n" + sourceURL + "\nMessage from server = "+message);
            setStatus(STATUS_ERROR);
            throw new IOException("Error from server :" +message +"\nNo size returned from server");
        }
        if(this.size!=-1 && this.size!=sizeFromServer){
            setStatus(STATUS_ERROR);
            throw new IOException("The size returned by the server doesn't match the expected size.");
        }
        this.size = sizeFromServer;
        InputStream is = con.getInputStream();
        cis = new CounterInputStream(is);

        //prepare message digest
        Map<String,DigestInputStream>digestStreams = new HashMap<String,DigestInputStream>();
        InputStream lastInputStream = cis;
        {
            Iterator<Map.Entry<String, MessageDigest>> it = messageDigestMap.entrySet().iterator();
            while(it.hasNext()){
                Map.Entry<String, MessageDigest> entry = it.next();
                String alg = entry.getKey();
                DigestInputStream dis = new DigestInputStream(lastInputStream, entry.getValue());
                digestStreams.put(alg, dis);
                lastInputStream = dis;
            }
        }
        
        //download files
        setStatus(STATUS_DOWNLOADING);
        
        InputStream bis = new BufferedInputStream(lastInputStream,inputBufferSize);
        OutputStream os = new BufferedOutputStream(
                new FileOutputStream(this.getLocalAbsoluteFile()), outputBufferSize);
        byte[] internalBuffer = new byte[internalBufferSize];
        int c;
        while((c=bis.read(internalBuffer))!=-1){
           os.write(internalBuffer,0,c);
           this.setCounter(cis.getCounter());
           if(interrupt){
                is.close();os.close();
                throw new InterruptedException("download interrupted.");
           }
        }
        bis.close();
        os.close();
        
        //get message digests
        digestsMap = new HashMap<String,byte[]>();
        {
            Iterator<Map.Entry<String,DigestInputStream>> it = digestStreams.entrySet().iterator();
            while(it.hasNext()){
                Map.Entry<String,DigestInputStream> entry = it.next();
                String alg = entry.getKey();
                byte[] digest = entry.getValue().getMessageDigest().digest();
                digestsMap.put(alg, digest);
            }
        }
        setStatus(STATUS_DOWNLOADED);
    }
    
   public static class CounterInputStream extends FilterInputStream{

        private long counter=0;
      
        public CounterInputStream(InputStream is){
            super(is);
        }
        
        public void close() throws IOException{
            super.close();
        }

        public void mark(int readlimit){
            System.err.println("DownloaderProgressInputStream: mark unsupported.");
        }

        public boolean markSupported(){
            return false;
        }

        public int read() throws IOException{
            int n = super.read();
            if(n>=0){incrementCounter(1);}
            return n;
        }
        public int read(byte[] b) throws IOException{
            int result = super.read(b);
            if(result>=0){incrementCounter(result);}
            return result;
        }
        public int read(byte[] b, int off, int len) throws IOException{
            int result = super.read(b, off, len);
            if(result>=0){incrementCounter(result);}
            return result;
        }
        public void reset(){
            System.err.println("DownloaderFileInputStream: Reset unsupported.");
            return;
        }

        public long skip(long n) throws IOException{
            long result = super.skip(n);
            incrementCounter(result);
            return result;
        }

        public synchronized long getCounter(){
            return this.counter;
        }
        private synchronized void incrementCounter(long increment){
            counter+=increment;
        }
        
    }
}