/*
Copyright (c) 2006, CARDON DE LICHTBUER Rodolphe
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list 
of conditions and the following disclaimer. 

2. Redistributions in binary form must reproduce the above copyright notice, this 
list of conditions and the following disclaimer in the documentation and/or 
other materials provided with the distribution. 

3. The name of the author or contributors may not be used to endorse or promote 
products derived from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED 
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH 
DAMAGE.
 */

package be.cardon.sign.shared;

import be.cardon.utils.OperatingSystem;
import be.cardon.cryptoapi.provider.*;
import java.security.KeyStore;
import java.security.Provider;
import be.cardon.utils.Debug;
import java.util.List;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Iterator;
import java.io.*;
import java.security.cert.X509Certificate;
import java.security.cert.Certificate;
import java.security.PrivateKey;
import java.security.Key;
import java.security.UnrecoverableKeyException;
import be.cardon.sign.xmldsig.*;
import java.net.URI;
import java.util.Calendar;

/**
 *
 * @author Rodolphe
 */
public class KeyStoreUtils {
    
    /** Creates a new instance of KeyStoreUtils */
    public KeyStoreUtils() {
    }
   
    public static KeyStore getMSKeyStore() throws java.security.KeyStoreException{
        if(!OperatingSystem.isWindows()){
            throw new java.security.KeyStoreException("Microsoft Key Store is only available on Windows platform.");
        }
        Provider MSProvider = new CryptoAPIProvider();
        //Note: the provider doesn't have to be registered. (see API)
        try{
            KeyStore MSKeyStore = KeyStore.getInstance("CryptoAPI", MSProvider);
            MSKeyStore.load(null, null); //load 'My' store
            return MSKeyStore;
            
        }catch(Exception e){
            throw new java.security.KeyStoreException("Unable to open the Windows KeyStore. "+e.getMessage(), e);
        }
    }
    
    public static List<KeyStoreEntry> getTableEntries(KeyStore ks) throws java.security.KeyStoreException{
        
        try{
            if(ks==null){
                throw new NullPointerException("KeyStoreUtils: getTableEntries: Unexpected null pointer for keystore");
            }
            List<KeyStoreEntry> entries = new ArrayList<KeyStoreEntry>();
            for(Enumeration<String> it=ks.aliases();it.hasMoreElements();){
                String alias = it.nextElement();
                if(ks.isKeyEntry(alias)){
                    entries.add(new KeyStoreEntry(ks, alias));
                }
            }
            Debug.println("found "+entries.size()+" entrie(s)");
            return entries;
        }catch(Exception e){
            throw new java.security.KeyStoreException(e);
        }
    }
    
    /**Returns the .keystore file or null if does'nt exist.
     *@throws KeyStoreException if cannot read the file
     */
    public static File getDefaultJKSKeyStore() throws java.security.KeyStoreException{
        String homeDirectory = System.getProperty("user.home");
        String defaultJKSName = ".keystore";
        File defaultJKSStore = new File(homeDirectory + "/" + defaultJKSName);
        if(!defaultJKSStore.exists()){
            return null;
        }
        if(!defaultJKSStore.canRead()){
            throw new java.security.KeyStoreException("Cannot read default JKS file "+defaultJKSStore.getAbsolutePath());
        }
        return defaultJKSStore;
    }
    
    /**get and load JKS Store.
     *returns null if bad password or invalid format.
     *@throws KeyStoreException for other exceptions.
     */
    public static KeyStore getJKSStore(File JKSFile, char[] password) throws java.security.KeyStoreException{
        try{
            if(!JKSFile.canRead()){
                throw new java.security.KeyStoreException("Cannot read the file "+ JKSFile.getAbsolutePath());
            }
            KeyStore ks = KeyStore.getInstance("JKS");
            FileInputStream fis = new FileInputStream(JKSFile);
            if(password==null){
                ks.load(fis, null);
            }else{
                ks.load(fis, password);
            }
            return ks;
        }catch(IOException e){
            return null;
        }catch(Exception e){
            throw new java.security.KeyStoreException(e);
        }
    }

    public static KeyStore getPKCS12Store(File PKCS12File, char[] password) throws java.security.KeyStoreException{
        try{
            if(!PKCS12File.canRead()){
                throw new java.security.KeyStoreException("Cannot read the file "+ PKCS12File.getAbsolutePath());
            }
            KeyStore ks = KeyStore.getInstance("PKCS12");
            FileInputStream fis = new FileInputStream(PKCS12File);
            if(password==null){
                ks.load(fis, null);
            }else{
                ks.load(fis, password);
            }
            return ks;
        }catch(IOException e){
            return null;
        }catch(Exception e){
            throw new java.security.KeyStoreException(e);
        }
    }
    
    /**PKSC11, token name is only used to give a name to the provider( pkcs11-tokenname) MUST BE ONE WORD.*/
    public static KeyStore getPKCS11Store(File PKCS11LibFile, char[] password, String tokenName) throws java.security.KeyStoreException{
        try{
            if(!PKCS11LibFile.canRead()){
                throw new java.security.KeyStoreException("Cannot read the file "+ PKCS11LibFile.getAbsolutePath());
            }
            File PKCS11ConfigFile = File.createTempFile("pkcs11_config","conf");
            Debug.println("config temp file"+PKCS11ConfigFile.getAbsolutePath());
            BufferedWriter writer = new BufferedWriter(
                    new OutputStreamWriter(
                    new FileOutputStream(PKCS11ConfigFile)));
            writer.write(String.format("name = %s\n",tokenName));
            writer.write(String.format("library = %s\n",PKCS11LibFile.getPath()));
            writer.close();
            
            Provider PKCS11SunProv = new sun.security.pkcs11.SunPKCS11(PKCS11ConfigFile.getPath());
            KeyStore ks = KeyStore.getInstance("PKCS11",PKCS11SunProv);
            PKCS11ConfigFile.delete();
            if(password==null){
                ks.load(null, null);
            }else{
                ks.load(null, password);
            }
            return ks;

        }catch(IOException e){
            return null;
        }catch(Exception e){
            throw new java.security.KeyStoreException(e);
        }
    }
}
