/*
Copyright (c) 2006, CARDON DE LICHTBUER Rodolphe
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list 
of conditions and the following disclaimer. 

2. Redistributions in binary form must reproduce the above copyright notice, this 
list of conditions and the following disclaimer in the documentation and/or 
other materials provided with the distribution. 

3. The name of the author or contributors may not be used to endorse or promote 
products derived from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED 
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH 
DAMAGE.
 */

package be.cardon.sign.standalone;

import be.cardon.utils.OperatingSystem;
import be.cardon.cryptoapi.provider.*;
import java.security.KeyStore;
import java.security.Provider;
import be.cardon.utils.Debug;
import java.util.List;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Iterator;
import java.io.File;
import java.io.IOException;
import java.io.FileInputStream;
import java.io.FileOutputStream; 
import java.security.cert.X509Certificate;
import java.security.cert.Certificate;
import java.security.PrivateKey;
import java.security.Key;
import java.security.UnrecoverableKeyException;
import java.io.Writer;
import java.io.OutputStream;
import be.cardon.sign.xmldsig.*;
import java.net.URI;
import java.util.Calendar;
import be.cardon.sign.standalone.XMLSignerGUI;
import be.cardon.sign.shared.KeyStoreEntry;
import be.cardon.sign.*;

/**
 *
 * @author Rodolphe
 */


public class XMLSigner {
    
    public static int STATE_SELECT_FILES = 0;
    public static int STATE_SELECT_KEY = 1;
    public static int STATE_SIGNING = 2;
    public static int LAST_STATE = STATE_SIGNING;
    private int state;
    private XMLSignerGUI gui;
    private List<FileItem> inputFileItems = new java.util.ArrayList<FileItem>();
    private File signatureFile;
    private KeyStoreEntry signinKeyStoreEntry; // loaded key store, alias, provider, chain
    private PrivateKey signingPrivateKey;
    private Calendar date;
    
    
    /** MAIN STARTUP
     * @param args the command line arguments
     */
    public static void main(String[] args) {
        
        boolean debug=false;
        if(args.length>0){
            String opt = args[0];
            if(opt.equalsIgnoreCase("-d")){
                debug=true;
            }else{
                printUsage();
                System.exit(1);
            }
        }
        
        Debug.setDebug(debug);
        new XMLSigner();
    }
    public static void printUsage(){
        System.out.println("Usage : XMLSigner [options]");
        System.out.println("Options:");
        System.out.println("    -d : debug");
    }
    /** Creates a new instance of XMLSigner */
    public XMLSigner() {
        this.state = XMLSigner.STATE_SELECT_FILES;
        
        //thread safety
        javax.swing.SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                startGUI();
            }
        });
    }
    
    public void startGUI(){
        gui = new XMLSignerGUI(this);
    }
    public void setSignatureFile(File file){
        this.signatureFile = file;
    }
    public void setSigningPrivateKey(PrivateKey key){
        this.signingPrivateKey = key;
    }
    public void setSigningKeyStoreEntry(KeyStoreEntry kse){
        this.signinKeyStoreEntry= kse;
    }

    public static KeyStore getMSKeyStore() throws XMLSignerException{
        if(!OperatingSystem.isWindows()){
            throw new XMLSignerException("Microsoft Key Store is only available on Windows platform.");
        }
        Provider MSProvider = new CryptoAPIProvider();
        //Note: the provider doesn't have to be registered. (see API)
        try{
            KeyStore MSKeyStore = KeyStore.getInstance("CryptoAPI", MSProvider);
            MSKeyStore.load(null, null); //load 'My' store
            return MSKeyStore;
            
        }catch(Exception e){
            throw new XMLSignerException("Unable to open the Windows KeyStore. "+e.getMessage(), e);
        }
    }
    
    public static List<KeyStoreEntry> getTableEntries(KeyStore ks) throws XMLSignerException{
        
        try{
            if(ks==null){
                throw new XMLSignerException("XMLSigner: getTableEntries: Unexpected null pointer for keystore");
            }
            List<KeyStoreEntry> entries = new ArrayList<KeyStoreEntry>();
            for(Enumeration<String> it=ks.aliases();it.hasMoreElements();){
                String alias = it.nextElement();
                if(ks.isKeyEntry(alias)){
                    entries.add(new KeyStoreEntry(ks, alias));
                }
            }
            Debug.println("found "+entries.size()+" entrie(s)");
            return entries;
        }catch(Exception e){
            throw new XMLSignerException("Unable to list the keys in the keystores.", e);
        }
    }
    
    /**Returns the .keystore file or null if does'nt exist.
     *@throws XMLSignerException if cannot read the file
     */
    public static File getDefaultJKSKeyStore() throws XMLSignerException{
        String homeDirectory = System.getProperty("user.home");
        String defaultJKSName = ".keystore";
        File defaultJKSStore = new File(homeDirectory + "/" + defaultJKSName);
        if(!defaultJKSStore.exists()){
            return null;
        }
        if(!defaultJKSStore.canRead()){
            throw new XMLSignerException("Cannot read default JKS file "+defaultJKSStore.getAbsolutePath());
        }
        return defaultJKSStore;
    }
    
    /**get and load JKS Store.
     *returns null if bad password or invalid format.
     *@throws XMLSignerException for other exceptions.
     */
    public static KeyStore getJKSStore(File JKSFile, char[] password) throws XMLSignerException{
        try{
            if(!JKSFile.canRead()){
                throw new XMLSignerException("Cannot read the file "+ JKSFile.getAbsolutePath());
            }
            KeyStore ks = KeyStore.getInstance("JKS");
            FileInputStream fis = new FileInputStream(JKSFile);
            if(password==null){
                ks.load(fis, null);
            }else{
                ks.load(fis, password);
            }
            return ks;
        }catch(IOException e){
            return null;
        }catch(Exception e){
            throw new XMLSignerException(e);
        }
    }
    
    /** write a summary to given writer.*/
    public void summary(Writer writer){
        try{
        String text;
        
        text = "----- Files to sign -----\n\n";
        int nFiles = this.inputFileItems.size();
        text=text.concat("Number of files: "+nFiles+"\n\n");
        for(Iterator<FileItem> it=this.inputFileItems.iterator();it.hasNext();){
            FileItem fi = it.next();
            text=text.concat(fi.getFullPath()+"\n");
        }
        Debug.println(text);
        writer.append(text);
        
        
        text = "----- Signing key ------\n";
        text=text.concat("\nCertificate informations:");
        text=text.concat("\n- Subject DN: "+this.signinKeyStoreEntry.getSubject());
        text=text.concat("\n- Issuer DN: "+this.signinKeyStoreEntry.getIssuer());
        text=text.concat("\n- Expiration: "+this.signinKeyStoreEntry.getExpirationDate());
        text=text.concat("\n\nKeyStore provider informations:");
        text=text.concat("\n- Provider Name: "+this.signinKeyStoreEntry.getProvider().getName());
        text=text.concat("\n- Provider Version: "+this.signinKeyStoreEntry.getProvider().getVersion());
        text=text.concat("\n- KeyStore Type: "+this.signinKeyStoreEntry.getLoadedKeyStore().getType());
        Debug.println(text);
        writer.append(text);
        
        
        text = "\n\n----- Certificates chain ------\n";
        Certificate[] certs = this.signinKeyStoreEntry.getCertificateChain();
        for(int c = 0; c<certs.length; c++){
            X509Certificate cert = (X509Certificate)certs[c];
            text=text.concat("\nCertificate [" + c+ "]:");
            text=text.concat("\n- Subject =" +cert.getIssuerDN().toString());
        }
        X509Certificate lastCert = (X509Certificate)certs[certs.length-1];
        boolean rootCert=false;
        if(lastCert.getIssuerDN().equals(lastCert.getSubjectDN())){
            text=text.concat("\nLast certificate is a root certificate (Issuer DN = Subject DN)");
        }else{
            text=text.concat("\nWARNING! Last certificate is NOT a root certificate (Issuer DN != Subject DN)");
        }
        Debug.println(text);
        writer.append(text);
        
        
        }catch(Exception e){
            e.printStackTrace();
            try{
            String message = e.getClass().getCanonicalName() + ": " +e.getLocalizedMessage();
            writer.append("\nERROR in XMLSigner.summary(): "+message);
            }catch(Exception ex){
                e.printStackTrace();
            }
        }
    }
    
    /**Private key. return null if the password is wrong.
     *@throws XMLSignerException
     */
    public PrivateKey getPrivateKey(char[] pass) throws XMLSignerException{
        try{
            if(this.signinKeyStoreEntry==null){
                throw new XMLSignerException("Unexpected error in XMLSigner.sign() : signingKeyStoreEntry is null.");
            }
            KeyStore ks = this.signinKeyStoreEntry.getLoadedKeyStore();
            String alias = this.signinKeyStoreEntry.getAlias();
            PrivateKey key = getPrivateKey(ks, alias, pass);
            return key;
        }catch(Exception e){
            throw new XMLSignerException(e);
        }
    }
    
    
    
    /**Private key. return null if the password is wrong.
     *@throws XMLSignerException
     */
    private static PrivateKey getPrivateKey(KeyStore loadedKeyStore, String alias, char[] pass)
    throws XMLSignerException{
        try{
            Key key = loadedKeyStore.getKey(alias, pass);
            if(!(key instanceof PrivateKey)){
                throw new XMLSignerException("XMLSigner.getPrivateKey(): The key is not a private key!");
            }
            return (PrivateKey)key;
        }catch(UnrecoverableKeyException e){
            return null;
        }catch(java.lang.IllegalArgumentException e){
            //returned by JKS when pass is null.
            return null;
        }catch(Exception e){
            throw new XMLSignerException(e);
        }
    }
    
    public static Calendar getCurrentDate(){
        return Calendar.getInstance();
    }
    
    
    /**if date = null, calls getCurrentDate().*/
    public void sign(Writer writer)throws XMLSignerException{
        try{
            if(this.inputFileItems==null){
                 throw new XMLSignerException("Unexpected error in XMLSigner.sign() : inputFileItems is null.");
            }
            int nFiles = this.inputFileItems.size();
            if(nFiles==0){
                throw new XMLSignerException("No file to sign.");
            }
            File[] files = new File[this.inputFileItems.size()];            
            Iterator<FileItem> it = this.inputFileItems.iterator();
            for(int i = 0; i<nFiles; i++){
                FileItem item = it.next();
                files[i] = item.file;
            }
            
            File signatureFile = this.signatureFile;
            if(signatureFile==null){
                throw new XMLSignerException("First specify a signature file.");
            }
            URI baseURI = signatureFile.toURI();
            OutputStream fos = new FileOutputStream(signatureFile);
            
            PrivateKey privKey = this.signingPrivateKey;
            if(privKey==null){
                throw new XMLSignerException("Unexpected error in XMLSigner.sign() : private key is null.");
            }
            if(this.signinKeyStoreEntry==null){
                throw new XMLSignerException("Unexpected error in XMLSigner.sign() : signingKeyStoreEntry is null.");
            }
            Certificate[] chainArray = this.signinKeyStoreEntry.getCertificateChain();
            if(chainArray.length==0){
                throw new XMLSignerException("Certificate chain is null or empty.");
            }
            List<X509Certificate> chain = new ArrayList<X509Certificate>();
            for(int i = 0;i<chainArray.length;i++){
                if(!(chainArray[i] instanceof X509Certificate)){
                    throw new XMLSignerException("Certificate chain contains a no-X509 certificate.");
                }
                chain.add((X509Certificate)chainArray[i]);
            }

            Provider prov = this.signinKeyStoreEntry.getProvider();
            if(prov==null){
                throw new XMLSignerException("Provider is null.");
            }            
            
            Calendar dateSign; 
            if(this.date==null){
                dateSign = getCurrentDate();
            }else
                dateSign = this.date;
            
            sign(files, baseURI, privKey, chain, dateSign, fos, writer, prov);
            
        }catch(Exception e){
            throw new XMLSignerException(e);
        }
        
        
        
    }
    /**register the provider in first position and sign the given files.*/
    private static void sign(
            File[] files,
            URI baseURI,
            PrivateKey key,
            List<X509Certificate>chain,
            Calendar now,
            OutputStream os,
            Writer writer,
            Provider prov)throws XMLSignerException{
        try{
            //set provider in first position
            java.security.Security.removeProvider(prov.getName());
            java.security.Security.insertProviderAt(prov, 1);
            
            XMLSignerEngine.Sign2XML(files, baseURI, key, chain, now, os, writer);
            
        }catch(Exception e){
            throw new XMLSignerException(e);
        }
    }
    
    public java.util.List<FileItem> getInputFileItems(){
        return this.inputFileItems;
    }
    
    /**return the next state, or the current state if there is not next state.*/
    public int getNextState(){
        if(this.state == LAST_STATE){
            return LAST_STATE;
        }else{
            return this.state+1;
        }
    }
    
    /**return the next state, or the current state if there is not next state.*/
    public int getPreviousState(){
        if(this.state == 0){
            return 0;
        }else{
            return this.state - 1;
        }
    }
    
    public int getState(){
        return this.state;
    }
    
    public void setState(int state){
        this.state=state;
    }
    
}
