/*
Copyright (c) 2006, CARDON DE LICHTBUER Rodolphe
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list 
of conditions and the following disclaimer. 

2. Redistributions in binary form must reproduce the above copyright notice, this 
list of conditions and the following disclaimer in the documentation and/or 
other materials provided with the distribution. 

3. The name of the author or contributors may not be used to endorse or promote 
products derived from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED 
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH 
DAMAGE.
 */

package be.cardon.sign.xmldsig;

import  be.cardon.sign.xmldsig.dereferencer.DefaultDereferencer;

import javax.xml.crypto.*;
import javax.xml.crypto.dom.*;
import javax.xml.crypto.dsig.*;
import javax.xml.crypto.dsig.dom.*;
import javax.xml.crypto.dsig.keyinfo.*;
import javax.xml.crypto.dsig.spec.*;

import javax.xml.parsers.DocumentBuilderFactory;
import org.w3c.dom.Document; //interface
import org.w3c.dom.Text;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.w3c.dom.Node;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.Transformer;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import java.math.BigInteger;
import java.util.List;
import java.util.Calendar;
import java.util.ArrayList;
import java.util.Collections;
import java.security.*;
import java.security.cert.*;
import javax.security.auth.x500.X500Principal;


import java.io.OutputStream;
import java.io.File;
import java.io.IOException;
import be.cardon.utils.Debug;
import java.text.SimpleDateFormat;
import java.util.Iterator;
import java.net.URI;
import java.io.FileOutputStream;
import java.io.FileNotFoundException;
import java.io.Writer;
import java.io.PrintStream; 
import be.cardon.sign.xmldsig.dereferencer.implementations.ResolverLocalFilesystem;

/**
 *
 * @author Rodolphe
 */
public class XMLSignerEngine {
    
    // constant parameters :
    
    //digest : SHA-1 (no other method supported, MD5 is not recommanded by the 
    // XMLDSIG standard)
    public static final String jsr105ProviderPropertyName = "jsr105Provider";
    public static final String jsr105ProviderSun = "org.jcp.xml.dsig.internal.dom.XMLDSigRI";
    public static final String jsr105ProviderType = "DOM";
    public static final String canonicalizationMethod = CanonicalizationMethod.INCLUSIVE;
    public static final C14NMethodParameterSpec canonicalizationSpec = null; 
    public static final String DateDefinitionNS = "http://xml.cardon.be/date";
    
    
    
    
    /**Sign one file. Output in a XML Signature.
     * If no "jsr105Provider" property is defined, uses the "org.jcp.xml.dsig.internal.dom.XMLDSigRI" provider.
     * If you want to use another provider, set this property using {@link System#setProperty);
     * Create a reference with URI = simple name of the document (without path). For example : "mydoc.txt"
     * => The signature should be stored in the same directory !
     *@param file2sign The name of the file to sign.
     *@param key The private key used to sign the file.
     *@param chain The certificate chain associated with this key. This chain is stored in the signature.
     *@param os The OutputStream (typically a FileOutputStream) of the XML Signature.
     *@throws KeyException if the private key algorithm is not RSA or DSA.
     */
    public static void SignOneFile2XML(
            File file2sign, 
            PrivateKey key, 
            List<X509Certificate> chain,
            Calendar now,
            OutputStream os)
    throws 
            KeyException, 
            GeneralException, 
            MarshalException,
            XMLSignatureException{
        URI baseURI = file2sign.toURI();
        File[] files2Sign = {file2sign};
        Writer writer = XMLSignerEngine.StandardOutputWriter;
        Sign2XML( files2Sign, baseURI, key, chain, now, os, writer);
    }
        
    /**store the signature in signatureFile. signatureFile is used as baseURI.*/
    public static void Sign2XML(
            File[] files2sign,
            File signatureFile,
            PrivateKey key, 
            List<X509Certificate> chain,
            Calendar now,
            OutputStream os,
            Writer writer)
    throws KeyException, GeneralException, MarshalException, XMLSignatureException,
    FileNotFoundException{

        URI baseURI = signatureFile.toURI();
        FileOutputStream fos = new FileOutputStream(signatureFile);
        Sign2XML(files2sign, baseURI, key, chain, now, fos, writer);
    }
    
    
    public static void Sign2XML(
            File[] files2sign,
            URI baseURI,
            PrivateKey key, 
            List<X509Certificate> chain,
            Calendar now,
            OutputStream os,
            Writer writer)
    throws KeyException, GeneralException,  MarshalException, XMLSignatureException{
        try{
            writer.append("Starting...");
            // use SHA-1 has alg. (no other alg supported)
            String digestMethod=DigestMethod.SHA1;
            writer.append("\nDigestMethod: "+digestMethod);

            // check the private key alg, and set the signature method
            String signMeth; 
            String alg = key.getAlgorithm();
            
            writer.append("\nPrivate Key algorithm: "+alg);
            if(alg.equals("RSA")){
                signMeth = SignatureMethod.RSA_SHA1;
            }else if(alg.equals("DSA")){
                signMeth = SignatureMethod.DSA_SHA1;
            }else{
                String err = "Invalid private key alorithm ("+alg+"). Must be RSA or DSA.";
                writer.append("\n"+err);
                throw new KeyException(err);
            }

            //get the preferred XMLDSIG Provider.
            XMLSignatureFactory fac = getXMLSignatureFactory();

            //get the DOM DocumentBuilderFactory
            DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            dbf.setNamespaceAware(true); // support for XML namespace (must be set). => TO DO
            Document doc = dbf.newDocumentBuilder().newDocument(); //empty doc
            
            
            //list of reference
            List<Reference> refs = new ArrayList<Reference>();

            //reference to signatureTime object
            Reference signatureDateRef = fac.newReference(
                    "#TimeStamp",
                    fac.newDigestMethod(digestMethod, null),
                    Collections.singletonList(fac.newTransform(
                    CanonicalizationMethod.INCLUSIVE, 
                    (TransformParameterSpec)null)),
                    null, //SignatureProperties.TYPE
                    null);
            refs.add(signatureDateRef);
            
            //reference to files2sign
            writer.append("\nCreating References...");
            for(int fileCounter = 0; fileCounter < files2sign.length; fileCounter++){
                 File file2sign = files2sign[fileCounter];
                 
                //Debug.println("Current working directory : "+ currentWorkingDirectory);
                Debug.println("File to sign : "+ file2sign.getCanonicalPath());
                Debug.println("File to sign URI: "  +file2sign.toURI());
                 //reference to file2sign
                
                //to do : relativepath
                    String filename2sign = 
                            be.cardon.utils.URIUtilities.relativizeURI2(
                                baseURI,
                                file2sign.toURI()).toASCIIString();
                    Debug.println("decoded self URI path ="+new URI(filename2sign).getPath());
                    writer.append("\nAdd Reference : "+ filename2sign);
                    Reference fileRef = fac.newReference(
                        filename2sign,
                        fac.newDigestMethod(digestMethod, null));
                    
                    //for the future : predigested info
                    /*
                    newReference(String uri,
                                       DigestMethod dm,
                                       List transforms,
                                       String type,
                                       String id,
                                       byte[] digestValue)
                    */
                   refs.add(fileRef);
            }

            //SignedInfo
            CanonicalizationMethod canon = fac.newCanonicalizationMethod(
                    canonicalizationMethod,
                    canonicalizationSpec);
            SignedInfo si = fac.newSignedInfo(
                    canon, 
                    fac.newSignatureMethod(signMeth, null),
                    refs);
            
            //KeyInfo including certificates chain and Subject name of the 
            //end certificate that contains the public key for validation
            KeyInfoFactory kif = fac.getKeyInfoFactory();
            List<Object> x509dataContent = new ArrayList<Object>(chain);
            X509Certificate endCertificate = chain.get(0);
            x509dataContent.add(kif.newX509IssuerSerial(
                    endCertificate.getSubjectX500Principal().toString(), 
                    endCertificate.getSerialNumber()));
            X509Data x509data = kif.newX509Data(x509dataContent);
            
            KeyInfo ki = kif.newKeyInfo(Collections.singletonList(x509data));
            
            
            //XMLObject containing the signature time.
                        //create the document that will hold the signature
            
            //signature date
            String dateUTCMils = Long.toString(now.getTimeInMillis());
            String dateFriendly = new SimpleDateFormat("EEE, d MMM yyyy HH:mm:ss Z").format(now.getTime());
            String abc = Long.toString(now.getTimeInMillis());
            
            Element signDateElement1 = doc.createElement("FriendlyDate");
            signDateElement1.appendChild(
                    doc.createTextNode(dateFriendly));
            Element signDateElement2 = doc.createElement("UTCMillisecsFromEpoc");
            signDateElement2.appendChild(
                    doc.createTextNode(dateUTCMils));
            
            List<DOMStructure> signElems = new ArrayList<DOMStructure>();
            signElems.add(new DOMStructure(signDateElement1));
            signElems.add(new DOMStructure(signDateElement2));
            
            SignatureProperty property = fac.newSignatureProperty
              (signElems, "#Signature-1", "TimeStamp"); 
   
            XMLObject object = fac.newXMLObject(
                    Collections.singletonList(property),  
                    null,
                    null, 
                    null);
            
            //list of all XMLObjects
            List objects = Collections.singletonList(object);
            
            //Signature
            XMLSignature signature = fac.newXMLSignature(
                    si, //signed info
                    ki, //key info
                    objects,
                    "Signature-1", // Signature ID
                    null // SignatureValue ID
                    );
            
            //DOMSignContext
            DOMSignContext signContext = new DOMSignContext(key, doc);
            
            //Base URI (configuring the dereferencer)
            Debug.println("BaseURI = "+baseURI.toString());
            signContext.setBaseURI(baseURI.toString());
            
            //Dereferencer
            signContext.setURIDereferencer(new DefaultDereferencer(doc));
            //bug resolved => build-in dereferencer (resolver from apache security
            //doesn't decode encoded non-ASCII char => results in 'file not found' error
            
            //generated and sign the XMLSignature. The DOM document will contain 
            //the XML Signature if this method returns successfully.
            writer.append("\nGenerating and signing the XMLSignature...");
            signature.sign(signContext);
            writer.append("\nSaving signature to output stream...");
            //output
            TransformerFactory tf = TransformerFactory.newInstance();
            tf.newTransformer().transform(
                    new DOMSource(doc),
                    new StreamResult(os)
                   );
        }catch(Exception e){
            throw new GeneralException(e);
        }
    }
    
    /**
     *@DEPRECATED
     */
    public static SignatureStatus VerifierOneFileXMLSignature(File signature, File baseDirectory)
    throws GeneralException,
            IOException,
            java.security.KeyException
    {
        return VerifyXMLSignature(signature, baseDirectory);
    }
    
    
   /** Verifies the 'core' XML signature :
     * - check the digest value of each reference
     * - retrieves the public key in the certificates contained in the signature.
     * the XMLSignature MUST have 
     *      - a X509IssuerSerial element identifying the end user certificate.
     *      - OR a KeyValue element identifying the end user certificate.
     * - check the signature with the public key.
     *baseDirectory may be null (directory containing signature is used)
     *@throws CertificateChainNotFound
     *@throws java.security.KeyException
     * if the public key is not found
     *@throws GeneralException if there is zero or >1 Signature element(s).
     *@throws IOException if a file is unreadable
     *@throws KeyException if certificate with signing key not found.
     *@param signature Signature file.
     *@param baseDirectory base directory for dereferencer, may be null 
     *(uses signature directory).
     *@return a SignatureStatus with 
     *  - a boolean that indicates if the core verification succeed
     *  - a List with the reference URI and a boolean for each reference that
     *    indicates if the reference check succeed.
     *  - the certificate chain.
     */
    public static SignatureStatus VerifyXMLSignature(File signature, File baseDirectory)
    throws GeneralException,
            IOException,
            java.security.KeyException
    {
        try{

        // builder factory
        DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
        dbf.setNamespaceAware(true);
        Document doc = dbf.newDocumentBuilder().parse(signature);
        
        //Find signature element
        NodeList nl = doc.getElementsByTagNameNS(XMLSignature.XMLNS, "Signature");
        if(nl.getLength()==0){
            throw new GeneralException("Cannot find Signature element");
        }else if(nl.getLength()>1){
            //throw new GeneralException("This document containes more than one " +
            //        "Signature element. This case is not yet implemented.");
            System.err.println("This document containes more than one" +
                    "Signature element. This case is not yet implemented");
        }
        Node signatureNode = nl.item(0);
        DOMStructure signDOMStruct = new DOMStructure(signatureNode);

        //Create DOM XMLSignatureFactory
        XMLSignatureFactory fac = getXMLSignatureFactory();
        XMLSignature si = fac.unmarshalXMLSignature(signDOMStruct);
        
        //retrieves the certificate and selects public key
        KeyInfo keyinfo = si.getKeyInfo();
        List keycontent = keyinfo.getContent();
        Debug.println("XMLSigner Engine: key content num ="+keycontent.size());
        List<X509Certificate> certs = new ArrayList<X509Certificate>();
        X500Principal endCertPrincipal=null;
        BigInteger endCertSerial;
        PublicKey pubKey=null;
        X509Certificate signingCert=null;
        
        List x509content=null; 
        //retrieves the X509Data
        for(Iterator it=keycontent.iterator();it.hasNext();){
            Object obj = it.next();
            Debug.println("XMLSigner Engine: key content type = "+obj.getClass().getCanonicalName());
            if(obj instanceof KeyValue){
                pubKey = ((KeyValue)obj).getPublicKey();
                continue;
            }
            if(obj instanceof X509Data){
                x509content = ((X509Data) obj).getContent();
                continue;
            }
            Debug.println("XMLSigner Engine: WARNING: key content not used");
        }
        if(x509content==null){
            throw new GeneralException("X509Data element not found");
        }
        
        if(pubKey==null){
            //subject name and issuer OR public key
            for(Iterator it=x509content.iterator();it.hasNext();){
                Object obj = it.next();
                Debug.println("XMLSigner Engine: x509 content type = "+obj.getClass().getCanonicalName());
                if(obj instanceof X509IssuerSerial){    
                    X509IssuerSerial issuerSerial = (X509IssuerSerial) obj;
                    endCertPrincipal = new X500Principal(issuerSerial.getIssuerName());
                    endCertSerial = issuerSerial.getSerialNumber();
                    break;
                }
            }
            if(endCertPrincipal==null){
                  throw new java.security.KeyException("X509 IssuerSerial not found. " +
                   "Cannot identify the end certificate");
            }
        }

        //certificates and key       
        for(Iterator it=x509content.iterator();it.hasNext();){
            Object obj = it.next();
            if(obj instanceof X509Certificate){
                X509Certificate cert = (X509Certificate)obj;
                certs.add(cert);
                if(pubKey!=null & cert.getPublicKey().equals(pubKey)){
                    signingCert = cert;
                }else if(cert.getSubjectX500Principal().equals(endCertPrincipal)){
                    pubKey = cert.getPublicKey();
                    signingCert = cert;
                }
            }
        }        
        
        if(pubKey==null){
              throw new java.security.KeyException("End certificate corresponding to " +
               "the X509 IssuerSerial not found.");
        }
        
        //DOMValidateContext
        DOMValidateContext valContext = new DOMValidateContext(
                pubKey,
                signatureNode);
        
        if(baseDirectory!=null){
            valContext.setBaseURI(baseDirectory.toURI().toString());
        }else{
            valContext.setBaseURI(signature.toURI().toString());
        }
        
        //Dereferencer
        valContext.setURIDereferencer(new DefaultDereferencer(doc));
        //bug resolved => build-in dereferencer (resolver from apache security
        //doesn't decode encoded non-ASCII char => results in 'file not found' error
       
        //core validity
        SignatureStatus ss = new SignatureStatus(); 
        ss.signingCertificate = signingCert;
        boolean coreValidity = si.validate(valContext);
        ss.coreValid = coreValidity;
        boolean sv;
        List<ReferenceStatus> refsStatus = new ArrayList<ReferenceStatus>();
        
        if(coreValidity ==false){
            ss.signatureValid = si.getSignatureValue().validate(valContext);
        }else{
            ss.signatureValid = true;
        }
        Iterator i = si.getSignedInfo().getReferences().iterator();
        for(;i.hasNext();){
            Object obj = i.next();
            if(!(obj instanceof Reference)){
                throw new GeneralException("Incorrect 'References' element : all children" +
                        "must be 'Reference'.");
            }
            Reference ref = (Reference)obj;
            ReferenceStatus rs = new ReferenceStatus();
            if(coreValidity == false){
                rs.RefValidDigest = ref.validate(valContext);
            }else{
                rs.RefValidDigest = true;
            }
            rs.RefURI = ref.getURI();
            rs.refTranfs = ref.getTransforms();
            rs.refType = ref.getType();
            refsStatus.add(rs);
        }
        ss.refsStatus = refsStatus;
        ss.certs = certs;
        return ss;
        
        }catch(javax.xml.parsers.ParserConfigurationException e){
            throw new GeneralException(e);
        }catch(org.xml.sax.SAXException e){
            throw new GeneralException(e);
        }catch(ClassNotFoundException e){
            throw new GeneralException(e);
        }catch(InstantiationException e){
            throw new GeneralException(e);
        }catch(IllegalAccessException e){
            //when an application tries to reflectively create an instance 
            throw new GeneralException(e);
        }catch(javax.xml.crypto.MarshalException e){
            //unmarshalXMLSignature
            throw new GeneralException(e);
        }catch(XMLSignatureException e){
            throw new GeneralException(e);
        }
    }

        
    /**Get the preferred XMLDSIG Provider.*/
    public static XMLSignatureFactory getXMLSignatureFactory()
    throws
         ClassNotFoundException,
         InstantiationException,
         IllegalAccessException
    {
            String providerName = System.getProperty(
                    jsr105ProviderPropertyName, 
                    jsr105ProviderSun);
            Provider jsr105Provider = (Provider) (Class.forName(providerName).newInstance());
            XMLSignatureFactory fac = XMLSignatureFactory.getInstance(
                    jsr105ProviderType,
                    jsr105Provider);
            return fac;
    }
    
    public static Writer StandardOutputWriter = new Writer() {
        
        /**Close the stream, flushing it first. */ 
        public void close(){ }
        /**Flush the stream. */
        public void flush(){ }         
        /**Write a portion of an array of characters. */
        public void write(char[] cbuf, int off, int len) {
            String text = String.copyValueOf(cbuf, off, len);
            System.out.print(text);
        }
    };
    
    public static Writer nullWriter = new Writer() {
        /**Close the stream, flushing it first. */ 
        public void close(){ }
        /**Flush the stream. */
        public void flush(){ }         
        /**Write a portion of an array of characters. */
        public void write(char[] cbuf, int off, int len) { }
    };

    
}
