
/*
 * Copyright  1999-2004 The Apache Software Foundation.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
package be.cardon.sign.xmldsig.dereferencer;


import be.cardon.utils.Debug;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import com.sun.org.apache.xml.internal.security.signature.XMLSignatureInput;
import javax.xml.crypto.URIReference;
import org.w3c.dom.Document;



/**
 * During reference validation, we have to retrieve resources from somewhere.
 * This is done by retrieving a Resolver. The resolver needs two arguments: The
 * URI in which the link to the new resource is defined and the BaseURI of the
 * file/entity in which the URI occurs (the BaseURI is the same as the SystemId.
 *
 * @author $Author: raul $
 */
public class ResourceResolver {

   /** Field _alreadyInitialized */
   static boolean _alreadyInitialized = false;

   /** these are the system-wide resolvers */
   static List<ResourceResolver> _resolverVector = null;

   /** Field _individualResolverVector */
//   List<ResourceResolver> _individualResolverVector = null;
// NOT USED
   
   /** Field transformSpi */
   protected ResourceResolverSpi _resolverSpi = null;

   /**
    * Constructor ResourceResolver
    *
    * @param className
    * @throws ClassNotFoundException
    * @throws IllegalAccessException
    * @throws InstantiationException
    */
   private ResourceResolver(String className)
           throws ClassNotFoundException, IllegalAccessException,
                  InstantiationException {
      this._resolverSpi =
         (ResourceResolverSpi) Class.forName(className).newInstance();
   }

   /**
    * Constructor ResourceResolver
    *
    * @param resourceResolver
    */
   public ResourceResolver(ResourceResolverSpi resourceResolver) {
      this._resolverSpi = resourceResolver;
   }

   /**
    * Method getInstance
    *
    * @param uri
    * @param BaseURI
    * @return the instnace
    *
    * @throws ResourceResolverException
    */
   public static final ResourceResolver getInstance(URIReference uriRef, String BaseURI, Document sigDoc)
           throws ResourceResolverException {
      int length=ResourceResolver._resolverVector.size();
      for (int i = 0; i < length; i++) {
		  ResourceResolver resolver =
            (ResourceResolver) ResourceResolver._resolverVector.get(i);
         

         Debug.println("check resolvability by class " + resolver._resolverSpi.getClass().getCanonicalName());

         if ((resolver != null) && resolver.canResolve(uriRef, BaseURI, sigDoc)) {
            Debug.println("can be resolved by class " + resolver._resolverSpi.getClass().getName());
            return resolver;
         }
      }

      Object exArgs[] = {uriRef, BaseURI};

      throw new ResourceResolverException("utils.resolver.noClass", exArgs,
                                          uriRef, BaseURI);
   }
   /**
    * Method getInstance
    *
    * @param uri
    * @param BaseURI
    * @param individualResolvers
    * @return the instance
    *
    * @throws ResourceResolverException
    */
   public static final ResourceResolver getInstance(
           URIReference uriRef, String BaseURI, Document sigDoc, List individualResolvers)
              throws ResourceResolverException {
      Debug.println("I was asked to create a ResourceResolver and got " + individualResolvers.size());
      Debug.println(" extra resolvers to my existing " + ResourceResolver._resolverVector.size() + " system-wide resolvers");
      

      // first check the individual Resolvers
	  int size=0;
      if ((individualResolvers != null) && ((size=individualResolvers.size()) > 0)) {
         for (int i = 0; i < size; i++) {
            ResourceResolver resolver =
               (ResourceResolver) individualResolvers.get(i);

            if (resolver != null) {
               String currentClass = resolver._resolverSpi.getClass().getName();
               Debug.println("check resolvability by class " + currentClass);

               if (resolver.canResolve(uriRef, BaseURI, sigDoc)) {
                  return resolver;
               }
            }
         }
      }

	  return getInstance(uriRef,BaseURI, sigDoc);
   }

   /**
    * The init() function is called by org.apache.xml.security.Init.init()
    */
   public static void init() {

      if (!ResourceResolver._alreadyInitialized) {
         ResourceResolver._resolverVector = new ArrayList<ResourceResolver>(10);
         _alreadyInitialized = true;
      }
   }

   /**
    * Method register
    *
    * @param className
    */
   public static void register(String className) {
	ResourceResolver resolver = null;
        try {
           resolver = new ResourceResolver(className);
		   ResourceResolver._resolverVector.add(resolver);
        } catch (Exception e) {
			System.err.println("Error loading resolver " + className +" disabling it");
        }
   }

   /** FIXED by CARDON
    * Method registerAtStart
    *
    * @param className
    */
   public static void registerAtStart(String className) {
      ResourceResolver resolver = null;
      try {
           resolver = new ResourceResolver(className);
		   ResourceResolver._resolverVector.add(0, resolver);
        } catch (Exception e) {
			System.err.println("Error loading resolver " + className +" disabling it");
        }
   }


   /**
    * Method resolve
    *
    * @param uri
    * @param BaseURI
    * @return the resource
    *
    * @throws ResourceResolverException
    */
   public static XMLSignatureInput resolveStatic(URIReference uriRef, String BaseURI, Document sigDoc)
           throws ResourceResolverException {
     Debug.println("ResourceResolver.resolveStatic(): " + uriRef.getURI());
      ResourceResolver myResolver = ResourceResolver.getInstance(uriRef, BaseURI, sigDoc);

      return myResolver.resolve(uriRef, BaseURI, sigDoc);
   }

   /**
    * Method resolve
    *
    * @param uri
    * @param BaseURI
    * @return the resource
    *
    * @throws ResourceResolverException
    */
   public XMLSignatureInput resolve(URIReference uriRef, String BaseURI, Document sigDoc)
           throws ResourceResolverException {
      return this._resolverSpi.engineResolve(uriRef, BaseURI, sigDoc);
   }

   /**
    * Method setProperty
    *
    * @param key
    * @param value
    */
   public void setProperty(String key, String value) {
      this._resolverSpi.engineSetProperty(key, value);
   }

   /**
    * Method getProperty
    *
    * @param key
    * @return the value of the property
    */
   public String getProperty(String key) {
      return this._resolverSpi.engineGetProperty(key);
   }

   /**
    * Method addProperties
    *
    * @param properties
    */
   public void addProperties(Map properties) {
      this._resolverSpi.engineAddProperies(properties);
   }

   /**
    * Method getPropertyKeys
    *
    * @return all property keys.
    */
   public String[] getPropertyKeys() {
      return this._resolverSpi.engineGetPropertyKeys();
   }

   /**
    * Method understandsProperty
    *
    * @param propertyToTest
    * @return true if the resolver understands the property
    */
   public boolean understandsProperty(String propertyToTest) {
      return this._resolverSpi.understandsProperty(propertyToTest);
   }

   /**
    * Method canResolve
    *
    * @param uri
    * @param BaseURI
    * @return true if it can resolve the uri
    */
   private boolean canResolve(URIReference uriRef, String BaseURI, Document sigDoc) {
      return this._resolverSpi.engineCanResolve(uriRef, BaseURI, sigDoc);
   }
}
