/*
Copyright (c) 2006, CARDON DE LICHTBUER Rodolphe
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list 
of conditions and the following disclaimer. 

2. Redistributions in binary form must reproduce the above copyright notice, this 
list of conditions and the following disclaimer in the documentation and/or 
other materials provided with the distribution. 

3. The name of the author or contributors may not be used to endorse or promote 
products derived from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED 
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH 
DAMAGE.
 */

package be.cardon.utils;


import java.util.List;
import java.util.ArrayList;
import java.io.File;
import java.util.Iterator;
import java.net.URI;

/**URI Path object used with URI Utilities.
 *String to parse MUST have / as separator.
 *
 * @author Rodolphe
 */
public class Path {
    
    static boolean isCaseSensitive= be.cardon.utils.OperatingSystem.isFileCaseSensitive();
    List<String> dirs;
    String filename;
    boolean isAbsolute;
    
    /**create a new instance.*/
    public Path(List<String> dirs, String filename, boolean isAbsolute){
        if(dirs==null){
            throw new java.lang.NullPointerException();
        }
        this.dirs = dirs;
        this.filename = filename;
        this.isAbsolute = isAbsolute;
    }

    /** Parse path and creates a new instance of URI Path. The string must have '/' separator.
     */
     public static Path parse(String path) throws PathParserException{
         String separator = "/";
         return parse(path, separator); 
     }

     
    public static String getSystemSeparator(){
        return be.cardon.utils.OperatingSystem.getSeparator();
    }
            
    public static Path parse(String path, String separator) throws PathParserException{
        
        //cannot mix / and \ separator
        String invalidSeparator;
        if(separator.equals("/")){
            invalidSeparator = "\\";
        }else if(separator.equals("\\")){
            invalidSeparator = "/";
        }else{
            throw new PathParserException("Invalid separator: "+separator);
        }
        
        
        boolean isAbsolute=false;
        List<String> dirs = new ArrayList<String>();
        String filename=null;
        
        //work on a copy of path
        String p = new String(path);
        
        if(p.indexOf(invalidSeparator)>-1){
            throw new PathParserException(
                    "invalid character "+invalidSeparator+" at position "+path.indexOf(invalidSeparator) 
                    + "in: "+path);
        }
        
        //trivial case
        if(p.equals(".")){
            return new Path(dirs, null, false);
        }
        
        //absolute path if contains ':' (windows) or if the first carachter is '/' (unix)
        if(p.indexOf(separator)==0 || p.indexOf(":")!=-1){
            isAbsolute = true;
            
            int twoPoints = p.indexOf(":");
            if(twoPoints != -1 && twoPoints + 1< p.length() ){
                if(!p.substring(twoPoints+1, twoPoints+2).equals(separator)){
                    throw new PathParserException(separator +" expected after ':' in "+p);
                }
            }else  if(twoPoints != -1 && twoPoints + 1>= p.length() ){
                //resolve bug when path is trivial like "c:"
                return new Path(
                        java.util.Collections.singletonList(new String(p)),
                        null,  //no filename
                        true); //absolute path
            }
            
        }
        
        
        //remove the first '/' if any (unix absolute dir)
        if(p.indexOf(separator)==0){
            p = p.substring(1);
        }
        
        //retrieves the directories.
        int nextSeparator = -2;
        while(nextSeparator != -1){
            nextSeparator = p.indexOf(separator);
            if(nextSeparator != -1){
                String dir = p.substring(0, nextSeparator);
                if(!dir.equals(".") && dir.length()>0){
                    dirs.add(dir);
                }
                p = p.substring(nextSeparator+1);
            }
        }
        if(p.length()>0){
            filename = new String(p);
        }
        
        return new Path(dirs, filename, isAbsolute);
    }
    
    /**Returns true if the path is a file.*/
    public boolean isFile(){
        return (filename!=null);
    }
    
    /**Returns true if the path is a directory.*/
    public boolean isDirectory(){
        return (filename==null);
    }
    
    /**returns a list with the sequence of directory.
     *Can returns an empty list but never null.
     */
    public List<String> getDirectoriesList(){
        return new ArrayList<String>(this.dirs);
    }
    
    /**returns an array with the sequence of directory.
     *returns an empty string array, but never null
     */
    public String[] getDirectoriesArray(){
        return listStringToArrayString(getDirectoriesList());
    }
    
    public void setDirectories(List<String> dirs, boolean isAbsolute){
        if(dirs==null){
            throw new java.lang.IllegalArgumentException("dirs is null");
        }
        this.dirs = dirs;
        this.isAbsolute = isAbsolute;
    }
    
    /**returns the filename or null if the filenamt is a directory.*/
    public String getFilename(){
        if(this.filename==null){
            return null;
        }
        return new String(this.filename);
    }
    
    /**set the filename. The given string may be null. In that case, the path
     *object becomes a directory.*/
    public void setFilename(String filename){
        this.filename = filename;
    }
    
    public boolean isAbsolutePath(){
        return this.isAbsolute;
    }
    
    /**Returns true if the given Path is included in the Path (or is the same path).
     *If the current object is a file, only the parent directory is used
     */
    public boolean isChild(Path path){
        if(path==null){
            throw new NullPointerException("path");
        }
        Path relativePath = this.relativize(path);
        if(relativePath.isAbsolutePath()){
            Debug.println("Path: isChile: relativePath is absolute");
            return false;
        }
        Path normRelativePath = relativePath.normalize();
        int size = normRelativePath.dirs.size();
        if(size==0){
            //same path
            return true;
        }
        if(normRelativePath.dirs.get(0).equals("..")){
            Debug.println("Path: isChid: ../");
            return false;
        }else{
            return true;
        }
    }
    
    private static String[] listStringToArrayString(List<String> list){
        String[] array = new String[list.size()];
        int c=0;
        for(Iterator<String> it = list.iterator();it.hasNext();){
            array[c]= it.next();
            c++;
        }
        return array;
    }
    
    public String toString(){
        String separator = "/";
        return this.toString(separator);
    }
    public String toString(String separator){
        String s = new String();
        
        //trivial case "./";
        if(!this.isAbsolute && this.dirs.isEmpty() && this.filename==null){
            return "."+separator;
        }
        
        boolean isUnixAbsolute=false;
        if(this.isAbsolute){
            if(!this.dirs.isEmpty()){
                String firstDir = this.dirs.get(0);
                if(firstDir.indexOf(":")== -1){
                    isUnixAbsolute=true;
                }
            }else{
                isUnixAbsolute=true;
            }
        }
        if(isUnixAbsolute){
            s = s + separator;
        }
        for(Iterator it = dirs.iterator(); it.hasNext();){
            s = s + it.next() + separator;
        }
        if(filename!=null){
            s = s + filename;
        }
        
        return s;
    }
    
    public Path normalize(){
        try{
            
            
            //work on a copy of dirs
            List<String> dirs = new ArrayList(this.dirs);
            
            //remove '.'
            for(Iterator<String> it = dirs.iterator(); it.hasNext();){
                if(it.next().equals(".")){
                    it.remove();
                }
            }
            
            //remove ".." and parent, if not preceded by ".."
            boolean mustContinue=true;
            while(mustContinue){
                mustContinue = false;
                for(int index=1;index<dirs.size();index++){
                    if(dirs.get(index).equals("..")
                    && !dirs.get(index-1).equals("..")){
                        dirs.remove(index-1);
                        dirs.remove(index-1);
                        mustContinue = true;
                    }
                }
            }
            return new Path(dirs, this.filename, this.isAbsolute);
            
        }catch(Exception e){
            //on error, return the same path, and print error to standard error.
            e.printStackTrace();
            return this;
        }
    }
    
    /**returns toRelativizePath if not absolute path...
     *@throws NullPointerException if toRelativizePath is null
     */
    public Path relativize(Path toRelativizePath){
        
        if(toRelativizePath==null){
            throw new NullPointerException("toRelativizePath");
        }
        
        //check that the two paths are absolute, otherwise return toRelativizePath
        if(!this.isAbsolute || !toRelativizePath.isAbsolutePath()){
            return toRelativizePath;
        }
        
        //normalize the two paths
        Path basePathNorm = this.normalize();
        Path toRelPathNorm =  toRelativizePath.normalize();
        
        //get directories list (copy of)
        List<String> basePathDirs = basePathNorm.getDirectoriesList();
        List<String> toRelPathDirs = toRelPathNorm.getDirectoriesList();
        
        //suppress the common directories in the list
        Iterator<String> baseIt = basePathDirs.iterator();
        Iterator<String> toRelIt = toRelPathDirs.iterator();
        while(baseIt.hasNext() && toRelIt.hasNext()){
            String dir = baseIt.next();
            String dir2 = toRelIt.next();
            boolean dirEqual;
            if(isCaseSensitive){
                dirEqual=dir.equals(dir2);
            }else{
                dirEqual=dir.equalsIgnoreCase(dir2);
            }
            if(dirEqual){
                baseIt.remove();
                toRelIt.remove();
            }else{
                break;
            }
        }
        
        for(int i=0;i<basePathDirs.size();i++){
            toRelPathDirs.add(0, "..");
        }
        
        return new Path(toRelPathDirs,
                toRelativizePath.getFilename(),
                false);
        
    }
}
