/*
Copyright (c) 2006, CARDON DE LICHTBUER Rodolphe
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list 
of conditions and the following disclaimer. 

2. Redistributions in binary form must reproduce the above copyright notice, this 
list of conditions and the following disclaimer in the documentation and/or 
other materials provided with the distribution. 

3. The name of the author or contributors may not be used to endorse or promote 
products derived from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED 
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH 
DAMAGE.
 */

package be.cardon.utils;

import java.security.KeyStore;
import java.security.KeyStoreSpi;
import java.security.cert.Certificate;
import java.security.Key;
import java.util.List;
import java.util.ArrayList;
import java.io.InputStream;
import java.util.Enumeration;
import java.util.Vector;
import java.security.KeyStoreException;
import java.util.Date;
import java.util.Iterator;
import java.io.OutputStream;



/**Basic keyStore. This keyStore is created from given certificates and keys
 * The aliases are automatically generated with a simple number.
 * This class is usefull when a keystore is requested and you have a list or array
 * of certs.
 *
 * @author Rodolphe
 */
public class SimpleKeyStore extends KeyStoreSpi{
    
    private List<CertificateEntry> certEntries;
    private List<KeyEntry> keyEntries;

    /** Creates a new instance of SimpleKeyStore */
    public SimpleKeyStore() {
        this.certEntries = new ArrayList<CertificateEntry>();
        this.keyEntries = new ArrayList<KeyEntry>();
        
    }
    
    
    public Enumeration<String> engineAliases(){
        Vector<String> aliases = new Vector<String>();
        for(Iterator<CertificateEntry> it = this.certEntries.iterator();it.hasNext();){
            aliases.add(it.next().getAlias());
        }
        for(Iterator<KeyEntry> it = this.keyEntries.iterator();it.hasNext();){
            aliases.add(it.next().getAlias());
        }
        return aliases.elements();
    }
    
    public boolean engineContainsAlias(String alias){
        for(Iterator<CertificateEntry> it = this.certEntries.iterator();it.hasNext();){
            if(it.next().getAlias().equals(alias)){
                return true;
            }
        }
        for(Iterator<KeyEntry> it = this.keyEntries.iterator();it.hasNext();){
            if(it.next().getAlias().equals(alias)){
                return true;
            }
        }
        return false;
    }
    public void engineDeleteEntry(String alias)throws KeyStoreException {
        for(Iterator<CertificateEntry> it = this.certEntries.iterator();it.hasNext();){
            if(it.next().getAlias().equals(alias)){
                it.remove();
                return;
            }
        }
        for(Iterator<KeyEntry> it = this.keyEntries.iterator();it.hasNext();){
            if(it.next().getAlias().equals(alias)){
                it.remove();
                return;
            }
        }
        throw new KeyStoreException("Alias not found");
    }
    public Certificate engineGetCertificate(String alias){
        for(Iterator<CertificateEntry> it = this.certEntries.iterator();it.hasNext();){
            CertificateEntry ce = it.next();
            if(ce.getAlias().equals(alias)){
                return ce.getCertificate();
            }
        }
        return null;
    }
    
    public String engineGetCertificateAlias(Certificate cert){
        for(Iterator<CertificateEntry> it = this.certEntries.iterator();it.hasNext();){
            CertificateEntry ce = it.next();
            if(ce.getCertificate().equals(cert)){
                return ce.getAlias();
            }
            
        }
        return null;
    }
    
    public Certificate[] engineGetCertificateChain(String alias){
        for(Iterator<KeyEntry> it = this.keyEntries.iterator();it.hasNext();){
            KeyEntry ke = it.next();
            if(ke.getAlias().equals(alias)){
                return ke.getChain();
            }
        }
        return null;
    }
    
    public Date engineGetCreationDate(String alias) {
        //not implemented
        return null;
    }
    
    
    public Key engineGetKey(String alias, char[] password) {
        //pass not used
        for(Iterator<KeyEntry> it = this.keyEntries.iterator();it.hasNext();){
            KeyEntry ke = it.next();
            if(ke.getAlias().equals(alias)){
                return ke.getKey();
            }
        }
        return null;
    }
    
    public boolean engineIsCertificateEntry(String alias) {
        return (this.engineGetCertificate(alias)!=null);
    }
    public boolean engineIsKeyEntry(String alias){
        return (this.engineGetKey(alias, null)!=null);
    }
    
    public void engineLoad(InputStream stream, char[] password){
        return;
    }
    
    public void engineSetCertificateEntry(String alias, Certificate cert){
        if(this.engineContainsAlias(alias)){
            try{
            this.engineDeleteEntry(alias);
            }catch(Exception e){
                e.printStackTrace();
            }
        }
        CertificateEntry ce = new CertificateEntry(alias, cert);
        this.certEntries.add(ce);
    }
    
        public void engineSetKeyEntry(String alias, byte[] key, Certificate[] chain)
        throws KeyStoreException{
            throw new KeyStoreException("function engineSetKeyEntry not used in SimpleKeyStore");
            
        }
        
    public void engineSetKeyEntry(String alias, Key key, char[] password, Certificate[] chain) {
        //pass not used
        if(password!=null){
            System.err.println("password currently not used in SimpleKeyStore");
        }
        if(this.engineContainsAlias(alias)){
            try{
            this.engineDeleteEntry(alias);
            }catch(Exception e){
                e.printStackTrace();
            }
        }
        KeyEntry ke = new KeyEntry(alias, key, chain);
        this.keyEntries.add(ke);
    }
    public int engineSize(){
        return (this.certEntries.size() + this.keyEntries.size());
    }
    public void engineStore(OutputStream stream, char[] password) {
        System.err.println("engineStore not implemented in SimpleKeyStore");
        return;
    }
}
class KeyEntry{
    private String alias;
    private Key key;
    private Certificate[] chain;
    
    public KeyEntry(String alias, Key key, Certificate[] chain){
        this.alias = alias;
        this.key = key;
        this.chain = chain;
    }
    public String getAlias(){
        return new String(this.alias);
    }
    public Key getKey(){
        return this.key;
    }
    public Certificate[] getChain(){
        return this.chain;
    }
}


class CertificateEntry{
    private String alias;
    private Certificate cert;
    
    public CertificateEntry(String alias, Certificate cert){
        this.alias = alias;
        this.cert = cert;
    }
    public String getAlias(){
        return new String(this.alias);
    }
    public Certificate getCertificate(){
        return this.cert;
    }
}