/*
Copyright (c) 2006, CARDON DE LICHTBUER Rodolphe
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list 
of conditions and the following disclaimer. 

2. Redistributions in binary form must reproduce the above copyright notice, this 
list of conditions and the following disclaimer in the documentation and/or 
other materials provided with the distribution. 

3. The name of the author or contributors may not be used to endorse or promote 
products derived from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED 
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH 
DAMAGE.
 *
 * "or if the path of this URI is not a prefix of the path of the given URI"
 * relativize().
 * example: baseURI = "http://www.google.be/test/"
 *          URItoRelativize = "http://www.google.be/test2/index.html"
 * must return : "../test2/index.html"
 * but returns "http://www.google.be/test2/index.html"
 *
 * see : http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=6226081
 *       http://www.velocityreviews.com/forums/t145550-javaneturirelativizejavaneturi-not-really-working.html
 *
 */

package be.cardon.utils;

import java.net.URI;

/**
 *
 * @author Rodolphe
 */
public class URIUtilities {
    
    /** Creates a new instance of URIUtilities */
    public URIUtilities() {
    }
    
    /**Relativizes the given URI against this URI.
     *<p>The relativization of the given URI against this URI is computed as follows: </p>
     *<p>If either this URI or the given URI are opaque, or if the scheme and
     * authority components of the two URIs are not identical (<B> but not if the path
     * of this URI is not a prefix of the path of the given URI</B>), then the given
     * URI is returned. </p>
     *<p>Otherwise a new relative hierarchical URI is constructed with query
     * and fragment components taken from the given URI and with a path
     * component computed by removing this URI's path from the beginning of
     * the given URI's path. </p>
     *@param baseURI the base URI.
     *@param toRelativizeURI The URI to be relativized against the base URI.
     *@returns The resulting URI.
     *@throws NullPointerException If toRelativizeURI is null.
     */
    public static URI relativizeURI2(URI baseURI, URI toRelativizeURI)
    throws NullPointerException{
        
        if(toRelativizeURI==null){
            throw new NullPointerException("toRelativizeURI");
        }
        
        if(baseURI==null){
            Debug.println("UTIUtilities: baseURI null");
            return toRelativizeURI;
        }
        
        if(baseURI.isOpaque() || toRelativizeURI.isOpaque()){
            Debug.println("UTIUtilities: opaque");
            return toRelativizeURI;
        }
        
        if(!baseURI.isAbsolute() || !toRelativizeURI.isAbsolute()){
            Debug.println("URIUtilities: not absolute");
            return toRelativizeURI;
        }
        
        String baseURIAuth = baseURI.getAuthority();
        String toRelURIAth = toRelativizeURI.getAuthority();
        
        if(baseURIAuth!=null){
            if(toRelURIAth!=null){
                if(!baseURIAuth.equals(toRelURIAth)){
                    Debug.println("UTIUtilities: not same authority");
                    return toRelativizeURI;
                }
            }else{
                Debug.println("UTIUtilities: not same authority");
                return toRelativizeURI;
            }
            
        }else{
            //baseURIAuth ==null
            if(toRelURIAth!=null){
                Debug.println("UTIUtilities: not same authority");
                return toRelativizeURI;
            }
        }
        
        
        if(baseURI.getPath()==null || toRelativizeURI.getPath()==null){
            Debug.println("UTIUtilities: path null");
            return toRelativizeURI;
        }
        
        String baseURIPathString = baseURI.getPath();
        String toRelURIPathString = toRelativizeURI.getPath();
        
        try{
            Path baseURIPath = Path.parse(baseURIPathString);
            Path toRelURIPath = Path.parse(toRelURIPathString);
            Path RelativePath = baseURIPath.relativize(toRelURIPath);
            
            return new URI(
                    null,
                    null,
                    RelativePath.toString(),
                    toRelativizeURI.getQuery(),
                    toRelativizeURI.getFragment());
        }catch(Exception e){
            //abnormal exception
            e.printStackTrace();
            return toRelativizeURI;
        }
                
    }
    
    /** Return true if uri is child of the baseuri (one or multiple levels).
     * Example : http//www.site.com/data/mail/doc.txt is child
     *          of : http://www.site.com/data/mail/
     *          of : http//www.site.com/data/, 
     *          and of http//www.site.com/
     *if the base uri is a file, the parent directory is used.
     *Returns always false if baseuri is not absolute, or is opaque
     *Reutns always false if uri is opaque.
     */
    public static boolean isChild(URI baseuri, URI uri)
    throws NullPointerException{
        if(baseuri==null){
            Debug.println("UTIUtilities: isChild: baseuri is null");
            return false;
        }
        if(uri==null){
            throw new NullPointerException("uri");
        }
        if(!baseuri.isAbsolute() || baseuri.isOpaque()){
            Debug.println("baseuri not absolute or opaque");
            return false;
        }
        if(uri.isOpaque()){
             Debug.println("uri opaque");
            return false;
        }
        
        //relativize the uri
        URI relativeURI = URIUtilities.relativizeURI2(baseuri, uri);
        
       Debug.println("relative URI :" +relativeURI);
           
        if(relativeURI.isAbsolute()){
           
            Debug.println("relative URI is absolute"); 
           return false;
        }
        
        String pathString = relativeURI.getPath();
        try{
        Path basePath = Path.parse(baseuri.getPath());
        Path relPath = Path.parse(relativeURI.getPath());
        return basePath.isChild(relPath);
        
        }catch(PathParserException e){
            e.printStackTrace();
            return false;
        }
        
        
        
    }
}
